import assert from 'assert';
import { Store, DefaultOptions, HeadsCache } from './../store.mjs';
import { default as Cache } from '@dao-xyz/peerbit-cache';
import { Keystore } from "@dao-xyz/peerbit-keystore";
import { X25519PublicKey } from '@dao-xyz/peerbit-crypto';
import { AccessError } from "@dao-xyz/peerbit-crypto";
import { SimpleIndex } from './utils.mjs';
import { fileURLToPath } from 'url';
import path from 'path';
import { jest } from '@jest/globals';
const __filename = fileURLToPath(import.meta.url);
const __filenameBase = path.parse(__filename).base;
// Test utils
import { nodeConfig as config, startIpfs, stopIpfs, createStore } from '@dao-xyz/peerbit-test-utils';
import { delay, waitFor } from '@dao-xyz/peerbit-time';
import { Address } from './../io.mjs';
const API = 'js-ipfs';
describe(`addOperation`, function () {
    let ipfsd, ipfs, signKey, keystore, identityStore, store, cacheStore, senderKey, recieverKey, encryption;
    let index;
    jest.setTimeout(config.timeout);
    const ipfsConfig = Object.assign({}, config, {
        repo: 'repo-encryption' + __filenameBase + new Date().getTime()
    });
    beforeAll(async () => {
        identityStore = await createStore(path.join(__filename, 'identity'));
        cacheStore = await createStore(path.join(__filename, 'cache'));
        keystore = new Keystore(identityStore);
        signKey = await keystore.createEd25519Key();
        ipfsd = await startIpfs(API, ipfsConfig.daemon1);
        ipfs = ipfsd.api;
        index = new SimpleIndex();
        senderKey = await keystore.createEd25519Key();
        recieverKey = await keystore.createEd25519Key();
        encryption = {
            getEncryptionKeypair: () => Promise.resolve(senderKey.keypair),
            getAnyKeypair: async (publicKeys) => {
                for (let i = 0; i < publicKeys.length; i++) {
                    if (publicKeys[i].equals(await X25519PublicKey.from(senderKey.keypair.publicKey))) {
                        return {
                            index: i,
                            keypair: senderKey.keypair
                        };
                    }
                    if (publicKeys[i].equals(await X25519PublicKey.from(recieverKey.keypair.publicKey))) {
                        return {
                            index: i,
                            keypair: recieverKey.keypair
                        };
                    }
                }
            }
        };
    });
    afterAll(async () => {
        await store?.close();
        ipfsd && await stopIpfs(ipfsd);
        await identityStore?.close();
        await cacheStore?.close();
    });
    it('encrypted entry is appended known key', async () => {
        const data = { data: 12345 };
        let done = false;
        const onWrite = async (store, entry) => {
            try {
                const heads = store.oplog.heads;
                expect(heads.length).toEqual(1);
                assert(Address.isValid(store.address));
                assert.deepStrictEqual(entry.payload.getValue(), data);
                expect(store.replicationStatus.progress).toEqual(1n);
                expect(store.replicationStatus.max).toEqual(1n);
                assert.deepStrictEqual(index._index, heads);
                await delay(5000); // seems because write is async?
                store._cache.getBinary(store.localHeadsPath, HeadsCache).then(async (localHeads) => {
                    if (!localHeads) {
                        fail();
                    }
                    localHeads.heads[0].init({
                        encryption: store.oplog._encryption,
                        encoding: store.oplog._encoding
                    });
                    await localHeads.heads[0].getPayload();
                    assert.deepStrictEqual(localHeads.heads[0].payload.getValue(), data);
                    assert(localHeads.heads[0].equals(heads[0]));
                    expect(heads.length).toEqual(1);
                    expect(localHeads.heads.length).toEqual(1);
                    done = true;
                });
            }
            catch (error) {
                throw error;
            }
        };
        const cache = new Cache(cacheStore);
        const options = { ...DefaultOptions, resolveCache: () => Promise.resolve(cache), onUpdate: index.updateIndex.bind(index), encryption, onWrite };
        store = new Store({ name: 'name' });
        await store.init(ipfs, {
            ...signKey.keypair,
            sign: async (data) => (await signKey.keypair.sign(data))
        }, options);
        await store._addOperation(data, {
            reciever: {
                clock: recieverKey.keypair.publicKey,
                payload: recieverKey.keypair.publicKey,
                signature: recieverKey.keypair.publicKey
            }
        });
        await waitFor(() => done);
    });
    it('encrypted entry is append unkown key', async () => {
        const data = { data: 12345 };
        let done = false;
        const onWrite = (store, entry) => {
            const heads = store.oplog.heads;
            expect(heads.length).toEqual(1);
            assert(Address.isValid(store.address));
            assert.deepStrictEqual(entry.payload.getValue(), data);
            expect(store.replicationStatus.progress).toEqual(1n);
            expect(store.replicationStatus.max).toEqual(1n);
            assert.deepStrictEqual(index._index, heads);
            store._cache.getBinary(store.localHeadsPath, HeadsCache).then(async (localHeads) => {
                if (!localHeads) {
                    fail();
                }
                localHeads.heads[0].init({
                    encryption: store.oplog._encryption,
                    encoding: store.oplog._encoding
                });
                try {
                    await localHeads.heads[0].getPayload();
                    assert(false);
                }
                catch (error) {
                    expect(error).toBeInstanceOf(AccessError);
                }
                assert(localHeads.heads[0].equals(heads[0]));
                expect(heads.length).toEqual(1);
                expect(localHeads.heads.length).toEqual(1);
                done = true;
            });
        };
        const cache = new Cache(cacheStore);
        const options = { ...DefaultOptions, resolveCache: () => Promise.resolve(cache), onUpdate: index.updateIndex.bind(index), encryption, onWrite };
        store = new Store({ name: 'name' });
        await store.init(ipfs, {
            ...signKey.keypair,
            sign: async (data) => (await signKey.keypair.sign(data))
        }, options);
        const reciever = await keystore.createEd25519Key();
        await store._addOperation(data, {
            reciever: {
                clock: undefined,
                payload: reciever.keypair.publicKey,
                signature: reciever.keypair.publicKey,
            }
        });
        await waitFor(() => done);
    });
});
//# sourceMappingURL=encryption.test.js.map