import assert from 'assert';
import { Store, DefaultOptions } from './../store.mjs';
import { default as Cache } from '@dao-xyz/peerbit-cache';
import { Keystore } from "@dao-xyz/peerbit-keystore";
import { SimpleIndex } from './utils.mjs';
import { delay, waitFor } from '@dao-xyz/peerbit-time';
import { fileURLToPath } from 'url';
import { jest } from '@jest/globals';
import path from 'path';
const __filename = fileURLToPath(import.meta.url);
const __filenameBase = path.parse(__filename).base;
// Test utils
import { nodeConfig as config, testAPIs, startIpfs, stopIpfs, createStore } from '@dao-xyz/peerbit-test-utils';
Object.keys(testAPIs).forEach(API => {
    describe(`orbit-db - Replication Status (${API})`, function () {
        jest.setTimeout(config.timeout);
        let ipfsd, ipfs, signKey, identityStore, cache1, cache2;
        let store1;
        let store2;
        let index1;
        let index2;
        jest.setTimeout(config.timeout);
        const ipfsConfig = Object.assign({}, config, {
            repo: 'repo-add' + __filenameBase + new Date().getTime()
        });
        beforeAll(async () => {
            identityStore = await createStore(path.join(__filename, 'identity'));
            const keystore = new Keystore(identityStore);
            signKey = await keystore.createEd25519Key();
            ipfsd = await startIpfs('js-ipfs', ipfsConfig.daemon1);
            ipfs = ipfsd.api;
            cache1 = await createStore(path.join(__filename, 'cache1'));
            cache2 = await createStore(path.join(__filename, 'cache2'));
        });
        afterAll(async () => {
            await store1?.close();
            await store2?.close();
            ipfsd && await stopIpfs(ipfsd);
            await identityStore?.close();
            await cache1?.close();
            await cache2?.close();
        });
        beforeEach(async () => {
            await cache1.clear();
            await cache2.clear();
            index1 = new SimpleIndex();
            index2 = new SimpleIndex();
            store1 = new Store({ name: 'name' });
            await store1.init(ipfs, {
                ...signKey.keypair,
                sign: async (data) => (await signKey.keypair.sign(data))
            }, { ...DefaultOptions, resolveCache: () => Promise.resolve(new Cache(cache1)), onUpdate: index1.updateIndex.bind(index1) });
            store2 = new Store({ name: 'name' });
            await store2.init(ipfs, {
                ...signKey.keypair,
                sign: async (data) => (await signKey.keypair.sign(data))
            }, { ...DefaultOptions, resolveCache: () => Promise.resolve(new Cache(cache2)), onUpdate: index2.updateIndex.bind(index2) });
        });
        it('has correct status', async () => {
            // init
            assert.deepEqual(store1.replicationStatus, { progress: 0, max: 0 });
            // load
            await store1._addOperation('hello');
            await delay(3000); // <-- cache is async so we need a to wait a bit so the load actually catches the new entry
            await store1.close();
            await store1.load();
            await waitFor((() => store1.replicationStatus.progress === 1n && store1.replicationStatus.max === 1n));
            await store1.close();
            // close 
            await store1.close();
            assert.deepEqual(store1.replicationStatus, { progress: 0, max: 0 });
            // sync 
            await store1.load();
            await store1._addOperation('hello2');
            await store2.sync(store1._oplog.heads);
            await waitFor((() => store2.replicationStatus.progress === 2n && store2.replicationStatus.max === 2n));
            //  snapshot
            await store1._cache._store.open();
            await store1.saveSnapshot();
            await store1.close();
            await store1.loadFromSnapshot();
            assert.deepEqual(store1.replicationStatus, { progress: 2, max: 2 });
        });
    });
});
//# sourceMappingURL=replication-status.test.js.map