import PQueue from 'p-queue';
import { Log, ISortFunction, PruneOptions, LogOptions, Identity, CanAppend } from '@dao-xyz/ipfs-log';
import { Encoding, EncryptionTemplateMaybeEncrypted } from '@dao-xyz/ipfs-log';
import { Entry } from '@dao-xyz/ipfs-log';
import { Replicator } from './replicator.js';
import { ReplicationInfo } from './replication-info.js';
import Cache from '@dao-xyz/peerbit-cache';
import { Constructor } from '@dao-xyz/borsh';
import { IPFS } from 'ipfs-core-types';
import { PublicKeyEncryptionResolver } from "@dao-xyz/peerbit-crypto";
import { Address, Addressable } from './io.js';
import { SystemBinaryPayload } from '@dao-xyz/peerbit-bpayload';
import { EntryWithRefs } from './entry-with-refs.js';
export declare class CachedValue {
}
export declare class CID extends CachedValue {
    hash: string;
    constructor(opts?: {
        hash: string;
    });
}
export declare class UnsfinishedReplication extends CachedValue {
    hashes: string[];
    constructor(opts?: {
        hashes: string[];
    });
}
export declare class HeadsCache<T> extends CachedValue {
    heads: Entry<T>[];
    constructor(opts?: {
        heads: Entry<T>[];
    });
}
export interface IStoreOptions<T> {
    /**
     * f set to true, will throw an error if the database can't be found locally. (Default: false)
     */
    localOnly?: boolean;
    /**
     * The directory where data will be stored (Default: uses directory option passed to OrbitDB constructor or ./orbitdb if none was provided).
     */
    directory?: string;
    /**
     * Replicate the database with peers, requires IPFS PubSub. (Default: true)
     */
    replicate?: boolean;
    onClose?: (store: Store<T>) => void;
    onDrop?: (store: Store<T>) => void;
    onLoad?: (store: Store<T>) => void;
    onLoadProgress?: (store: Store<T>, entry: Entry<T>) => void;
    onWrite?: (store: Store<T>, _entry: Entry<T>) => void;
    onOpen?: (store: Store<any>) => Promise<void>;
    onReplicationQueued?: (store: Store<any>, entry: Entry<T>) => void;
    onReplicationProgress?: (store: Store<any>, entry: Entry<T>) => void;
    onReplicationComplete?: (store: Store<any>) => void;
    onReady?: (store: Store<T>) => void;
    encryption?: PublicKeyEncryptionResolver;
    maxHistory?: number;
    fetchEntryTimeout?: number;
    referenceCount?: number;
    replicationConcurrency?: number;
    syncLocal?: boolean;
    sortFn?: ISortFunction;
    prune?: PruneOptions;
    onUpdate?: (oplog: Log<T>, entries?: Entry<T>[]) => void;
}
export interface IInitializationOptions<T> extends IStoreOptions<T>, IInitializationOptionsDefault<T> {
    resolveCache: (address: Address) => Promise<Cache<CachedValue>> | Cache<CachedValue>;
}
interface IInitializationOptionsDefault<T> {
    maxHistory?: number;
    referenceCount?: number;
    replicationConcurrency?: number;
    typeMap?: {
        [key: string]: Constructor<any>;
    };
}
export declare const DefaultOptions: IInitializationOptionsDefault<any>;
export interface Initiable<T> {
    init?(ipfs: IPFS, identity: Identity, options: IInitializationOptions<T>): Promise<this>;
}
export interface Saveable {
    save(ipfs: any, options?: {
        format?: string;
        pin?: boolean;
        timeout?: number;
    }): Promise<Address>;
}
export declare class Store<T> extends SystemBinaryPayload implements Addressable, Initiable<T>, Saveable {
    name: string;
    _canAppend?: CanAppend<T>;
    _onUpdate?: (oplog: Log<T>, entries?: Entry<T>[]) => Promise<void> | void;
    _onUpdateOption?: (oplog: Log<T>, entries?: Entry<T>[]) => Promise<void> | void;
    _options: IInitializationOptions<T>;
    identity: Identity;
    remoteHeadsPath: string;
    localHeadsPath: string;
    snapshotPath: string;
    queuePath: string;
    manifestPath: string;
    initialized: boolean;
    address: Address;
    encoding: Encoding<T>;
    _ipfs: IPFS;
    _cache: Cache<CachedValue>;
    _oplog: Log<T>;
    _queue: PQueue<any, any>;
    _replicationStatus: ReplicationInfo;
    _stats: any;
    _replicator: Replicator<T>;
    _loader: Replicator<T>;
    _key: string;
    constructor(properties?: {
        name?: string;
        parent?: {
            name: string;
        };
    });
    setup(properties: {
        encoding: Encoding<T>;
        canAppend: CanAppend<T>;
        onUpdate: (oplog: Log<T>, entries?: Entry<T>[]) => void;
    }): void;
    init(ipfs: IPFS, identity: Identity, options: IInitializationOptions<T>): Promise<this>;
    updateStateFromLogs: (logs: Log<T>[]) => Promise<void>;
    get oplog(): Log<any>;
    get key(): string;
    get logOptions(): LogOptions<T>;
    /**
     * Returns the database's current replication status information
     * @return {[Object]} [description]
     */
    get replicationStatus(): ReplicationInfo;
    setIdentity(identity: Identity): void;
    set canAppend(canAppend: CanAppend<T> | undefined);
    get canAppend(): (CanAppend<T> | undefined);
    set onUpdate(onUpdate: (oplog: Log<T>, entries?: Entry<T>[]) => void);
    close(): Promise<void>;
    /**
     * Drops a database and removes local data
     */
    drop(): Promise<void>;
    load(amount?: number, opts?: {
        fetchEntryTimeout?: number;
    }): Promise<void>;
    /**
     *
     * @param heads
     * @returns true, synchronization resolved in new entries
     */
    sync(heads: (EntryWithRefs<T> | Entry<T>)[]): Promise<boolean>;
    loadMoreFrom(entries: (string[] | Entry<any>[] | EntryWithRefs<any>[])): void;
    get replicate(): boolean;
    getCachedHeads(): Promise<Entry<T>[]>;
    saveSnapshot(): Promise<import("ipfs-core-types/dist/src/root.js").AddResult[]>;
    loadFromSnapshot(): Promise<this>;
    _updateIndex(entries?: Entry<T>[]): Promise<void>;
    syncLocal(): Promise<void>;
    _addOperation(data: T, options?: {
        skipCanAppendCheck?: boolean;
        nexts?: Entry<T>[];
        onProgressCallback?: (any: any) => void;
        pin?: boolean;
        reciever?: EncryptionTemplateMaybeEncrypted;
    }): Promise<Entry<T>>;
    _recalculateReplicationProgress(): void;
    _recalculateReplicationMax(value: bigint | number): void;
    _recalculateReplicationStatus(maxTotal: bigint | number): void;
    _onLoadProgress(entry: Entry<any>): void;
    clone(): Store<T>;
    save(ipfs: IPFS, options?: {
        format?: string;
        pin?: boolean;
        timeout?: number;
    }): Promise<Address>;
    static load(ipfs: IPFS, address: Address, options?: {
        timeout?: number;
    }): Promise<Store<any>>;
}
export {};
